import React from 'react';
import {
  colorPaletteConfigSlice,
  querybuilderSlice,
  setQuerybuilderNodes,
  store,
} from '@graphpolaris/shared/lib/data-access/store';
import { GraphPolarisThemeProvider } from '@graphpolaris/shared/lib/data-access/theme';
import { configureStore } from '@reduxjs/toolkit';
import { Meta } from '@storybook/react';
import { Provider } from 'react-redux';
import QueryBuilder from './querybuilder';
import { Handles } from '../graph/reactflow/handles';
import { QueryMultiGraph } from '../graph/graphology/utils';

const Component: Meta<typeof QueryBuilder> = {
  component: QueryBuilder,
  title: 'QueryBuilder/Panel',
  decorators: [
    (story) => (
      <Provider store={store}>
        <GraphPolarisThemeProvider>
          <div
            style={{
              width: '100%',
              height: '95vh',
            }}
          >
            {story()}
          </div>
        </GraphPolarisThemeProvider>
      </Provider>
    ),
  ],
};

const graph = new QueryMultiGraph();
graph.addPill2Graphology(
  { type: 'entity', x: 100, y: 100, name: 'Entity Pill', fadeIn: false },
  '0'
);
graph.addPill2Graphology(
  { type: 'entity', x: 200, y: 200, name: 'Entity Pill 2', fadeIn: false },
  '10'
);
// graph.addNode('0', { type: 'entity', x: 100, y: 100, name: 'Entity Pill' });
graph.addPill2Graphology(
  {
    type: 'relation',
    x: 140,
    y: 140,
    name: 'Relation Pill',
    depth: { min: 0, max: 1 },
    fadeIn: false,
  },
  '1'
);
// addPill2Graphology(
//   '2',
//   {
//     type: 'attribute',
//     x: 170,
//     y: 160,
//     name: 'Attr string',
//     dataType: 'string',
//     matchType: 'EQ',
//     value: 'mark',
//     depth: { min: 0, max: 1 },
//     fadeIn: false,
//   },
//   graph
// );
// addPill2Graphology(
//   '3',
//   {
//     type: 'attribute',
//     x: 170,
//     y: 170,
//     name: 'Attr number',
//     dataType: 'float',
//     matchType: 'EQ',
//     depth: { min: 0, max: 1 },
//     fadeIn: false,
//   },
//   graph
// );
// addPill2Graphology(
//   '4',
//   {
//     type: 'attribute',
//     x: 130,
//     y: 120,
//     name: 'Attr bool',
//     dataType: 'bool',
//     matchType: 'EQ',
//     value: 'true',
//     depth: { min: 0, max: 1 },
//     fadeIn: false,
//   },
//   graph
// );
graph.addEdge('0', '1', {
  type: 'connection',
  sourceHandle: Handles.ToRelation,
  targetHandle: Handles.RelationLeft,
});
graph.addEdge('10', '1', {
  type: 'connection',
  sourceHandle: Handles.ToRelation,
  targetHandle: Handles.RelationRight,
});
// console.log(graph.getNodeAttributes('2'));
// graph.addEdge('2', '1', { type: 'attribute_connection' });
// graph.addEdge('3', '1', { type: 'attribute_connection' });
// graph.addEdge('4', '0', { type: 'attribute_connection' });
// graph.addEdge('0', '1', {
//   type: 'entity_relation',
//   targetHandle: handles.relation.fromEntity,
// });
// graph.addEdge('1', '0', {
//   type: 'entity_relation',
//   sourceHandle: handles.relation.entity,
// });

export const Simple = {
  args: {},
  play: async () => {
    store.dispatch(setQuerybuilderNodes(graph.export()));
  },
};

export default Component;
