import { SchemaFromBackend, Node, Edge } from '@graphpolaris/shared/lib/model/backend';
import Graph, { MultiGraph } from 'graphology';
import { Attributes } from 'graphology-types';

export class SchemaUtils {
  public static schemaBackend2Graphology(
    schemaFromBackend: SchemaFromBackend
  ): Graph {
    const { nodes, edges } = schemaFromBackend;
    // Instantiate a directed graph that allows self loops and parallel edges
    const schemaGraph = new MultiGraph({ allowSelfLoops: true });
    // console.log('parsing schema');
    // The graph schema needs a node for each node AND edge. These need then be connected

    nodes.forEach((node: Node) => {
      schemaGraph.addNode(node.name, {
        ...node,
        name: node.name,
        nodeCount: 0,
        summedNullAmount: 0,
        connectedRatio: 1,
        handles: ['entitySourceLeft', 'entityTargetLeft'],
        attributes: {
          ...node.attributes,
        },
        x: 0,
        y: 0,
      });
      // console.log(node, schemaGraph.nodes());
    });



    // The name of the edge will be name + from + to, since edge names are not unique
    edges.forEach((edge: Edge) => {
      const edgeID = [edge.name, '_', edge.from, edge.to].join(''); //ensure that all interpreted as string

      // This node is the actual edge
      schemaGraph.addDirectedEdgeWithKey(edgeID, edge.from, edge.to, {
        nodeCount: 0,
        summedNullAmount: 0,
        fromRatio: 0,
        toRatio: 0,
        handles: [],
        // width: 10,
        // height: 20,
        ...edge,
        attributes: edge.attributes,
      });
    });
    return schemaGraph;
  }
}

